/* eslint camelcase:0 */
import fetch from 'cross-fetch'
import { getHeaders } from 'githubauthreq'
import Pool from 'native-promise-pool'
const ghapi = process.env.GITHUB_API || 'https://api.github.com'
export function halt(milliseconds) {
	if (milliseconds < 1000) {
		console.warn(
			'halt accepts milliseconds, you may have attempted to send it seconds, as you sent a value below 1000 milliseconds'
		)
	}
	return new Promise(function (resolve, reject) {
		setTimeout(resolve, milliseconds)
	})
}
// =================================
// Fetch Repository
/**
 * Fetch data for Repostiory from Repository Name
 * @param repoFullName Repostory name, such as `'bevry/getrepos'`
 */
export async function getRepo(repoFullName) {
	const url = `${ghapi}/repos/${repoFullName}`
	const resp = await fetch(url, {
		headers: getHeaders(),
	})
	if (resp.status === 429) {
		// wait a minute
		console.warn(
			`${url} returned 429, too many requests, trying again in a minute`
		)
		await halt(60 * 1000)
		return getRepo(repoFullName)
	}
	const data = await resp.json()
	if (data && data.message) throw data.message
	if (!data || !data.full_name) throw new Error('response was not a repository')
	return data
}
/**
 * Fetch data for Repositories from Repository Names
 * @param repoFullNames Array of repository names, such as `['bevry/getcontributors', 'bevry/getrepos']`
 */
export async function getRepos(repoFullNames, concurrency = 0) {
	const pool = new Pool(concurrency)
	return await Promise.all(
		repoFullNames.map((repoFullName) => pool.open(() => getRepo(repoFullName)))
	)
}
// =================================
// Fetch from Search
/**
 * Fetch data for Repostiories from a Search, will iterate all subsequent pages
 * @param query The search query to send to GitHub, such as `@bevry/getcontributors @bevry/getrepos`
 */
export async function getReposFromSearch(query, opts = {}) {
	// defaults
	if (opts.page == null) opts.page = 1
	if (opts.pages == null) opts.pages = 10
	if (opts.size == null) opts.size = 100
	// fetch
	const url = `${ghapi}/search/repositories?page=${opts.page}&per_page=${
		opts.size
	}&q=${encodeURIComponent(query)}`
	const resp = await fetch(url, {
		headers: getHeaders(),
	})
	const data = await resp.json()
	if (data && data.message) throw data.message
	if (!data || !data.items || !Array.isArray(data.items))
		throw new Error('response was not the format we expected')
	if (data.items.length === 0) return []
	const within = opts.pages === 0 || opts.page < opts.pages
	if (data.items.length === opts.size && within)
		return data.items.concat(
			await getReposFromSearch(
				query,
				Object.assign({}, opts, { page: opts.page + 1 })
			)
		)
	return data.items
}
/**
 * Fetch data for Repostiories from Users
 * @param users Fetch repositories for these users, such as `['bevry', 'browserstate']`
 */
export async function getReposFromUsers(users, opts = {}) {
	const query = users.map((name) => `@${name}`).join('%20')
	return await getReposFromSearch(query, opts)
}
